import matplotlib.pyplot as plt
import seaborn as sns
import networkx as nx


def plot_causal_graph(model: "Model") -> None:
    """
    Plot the causal graph of the model.

    Args:
        model: Model, the model to plot the causal graph for

    Returns:
        None

    Example:
    ```
    plot_causal_graph(model)
    ```
    """
    G = nx.DiGraph(model.get_edges())
    pos = nx.layout.spring_layout(G)
    nx.draw(
        G,
        pos=pos,
        arrows=True,
        with_labels=True,
        node_size=2000,
        node_color="#D6D6D6",
        arrowsize=25,
    )


def plot_causal_attributions(model: "Model", outcome: str, normalise: bool = False, ax=None, **kwargs) -> None:
    """
    Plot the causal attribution of each node.

    Args:
        outcome: str, the node to calculate the causal attribution for
        normalise: bool, whether to normalise the causal attribution
        ax: matplotlib axis, the axis to plot the causal attribution on
        kwargs: additional keyword arguments to pass to the seaborn barplot function

    Returns:
        ax: matplotlib axis, the axis containing the causal attribution plot

    Example:
    ```
    ax = plot_causal_attributions(model, "y")
    ```
    """
    # Calculate the causal attribution
    causal_attributions = model.causal_attributions(
        outcome, normalise=normalise)

    # Plot the causal attribution
    if ax is None:
        fig, ax = plt.subplots()

    # Set bar colors. If normalised, use a single color, else determine based on value.
    if normalise:
        bar_colors = ['#119488'] * len(causal_attributions)
    else:
        bar_colors = causal_attributions.iloc[:, 0].map(
            lambda x: '#8AB17D' if x > 0 else '#B85450').tolist()

    sns.barplot(
        x=causal_attributions.iloc[:, 0], y=causal_attributions.index, hue=causal_attributions.index, palette=bar_colors, ax=ax, **kwargs)
    ax.set_title(f"Causal attribution of {outcome}")
    ax.set_xlabel("Causal attribution")
    ax.set_ylabel("Node")

    # Loop through the bars and place the text annotation inside or next to the bars
    for bar in ax.patches:
        bar_value = bar.get_width()
        text_x_position = bar.get_x() + bar.get_width()
        if bar_value < 0:  # Adjust text position for negative bars if necessary
            text_x_position = bar.get_x()
        ax.text(text_x_position, bar.get_y() + bar.get_height()/2,
                f'{bar_value:.2f}',
                va='center', ha='left' if bar_value < 0 else 'left', fontsize=9)

    return ax



import requests
import os
import toml
from pydantic import validate_call
from .data import Data
from .model import Model
from .utils import get_causadb_url, set_causadb_url


class CausaDB:
    """CausaDB client class to interact with the CausaDB system.
    """

    def __repr__(self) -> str:
        return f"<CausaDB client>"

    def __str__(self) -> str:
        return "CausaDB client"

    def __init__(self, token: str = None, custom_url: str = None) -> None:
        """Initializes the CausaDB client.

        Args:
            custom_url (str, optional): The URL of the CausaDB server. For custom deployments or development purposes. Defaults to None.
        """
        # If the token is not provided, try to load it from the config file
        if token is None:
            token = self._load_token()

        # If we now have a token, set it. If not, the user will have to use set_token.
        if token is not None:
            self.set_token(token)

        # If a custom URL is provided, set it
        if custom_url is not None:
            set_causadb_url(custom_url)

    def _load_token(self) -> str:
        """Load the token from the config file.

        Returns:
            str: The token secret.
        """
        dir_name = os.path.expanduser("~/.causadb")
        config_filepath = os.path.join(dir_name, "config.toml")

        if not os.path.exists(config_filepath):
            raise Exception("No token found")

        with open(config_filepath, "r") as f:
            config = toml.load(f)

        token_secret = config \
            .get("default", {}) \
            .get("token_secret", None)

        return token_secret

    @validate_call
    def set_token(self, token: str) -> None:
        """Set the token for the CausaDB client.

        Args:
            token (str): Token secret provided by CausaDB.

        Raises:
            Exception: If the token is invalid.
        """

        # Verify that the tokens are correct
        headers = {"token": token}
        response = requests.get(
            f"{get_causadb_url()}/account",
            headers=headers
        )

        # If the response is successful, set the tokens
        if response.status_code == 200:
            self.token = token
        else:
            raise Exception("Invalid token")

    @validate_call
    def create_model(self, model_name: str) -> Model:
        """Create a model and add it to the CausaDB system.

        Args:
            model_name (str): The name of the model.

        Returns:
            Model: The model object.
        """
        return Model(model_name, self)

    @validate_call
    def add_data(self, data_name: str) -> Data:
        """Add data to the CausaDB system.

        Args:
            data_name (str): The name of the data.

        Returns:
            Data: The data object.
        """
        return Data(data_name, self)

    @validate_call
    def get_model(self, model_name: str) -> Model:
        """Get a model by name.

        Args:
            model_name (str): The name of the model.

        Returns:
            Model: The model object.
        """
        headers = {"token": self.token}

        try:
            response = requests.get(
                f"{get_causadb_url()}/models/{model_name}", headers=headers
            ).json()
        except Exception as e:
            raise Exception(f"CausaDB server request failed: {e}")

        # If the model exists, return it
        model = Model(model_name, self)

        return model

    @validate_call
    def list_models(self) -> list[Model]:
        """List all models.

        Returns:
            list[Model]: A list of model objects.
        """
        headers = {"token": self.token}

        try:
            response = requests.get(
                f"{get_causadb_url()}/models", headers=headers
            ).json()
        except Exception as e:
            raise Exception(f"CausaDB server request failed: {e}")

        model_list = []
        for model_spec in response.get("models", []):
            model = Model(model_spec["name"], self)
            model_list.append(model)

        return model_list

    @validate_call
    def get_data(self, data_name: str) -> Data:
        """Get a data by name.

        Args:
            data_name (str): The name of the data.

        Returns:
            Data: The data object.
        """
        headers = {"token": self.token}
        try:
            response = requests.get(
                f"{get_causadb_url()}/data/{data_name}", headers=headers
            ).json()
        except Exception as e:
            raise Exception(f"CausaDB server request failed: {e}")

        data = Data(data_name, self)

        return data

    def list_data(self) -> list[Data]:
        """List all data.

        Returns:
            list[Data]: A list of data objects.
        """
        headers = {"token": self.token}

        try:
            response = requests.get(
                f"{get_causadb_url()}/data", headers=headers
            ).json()
        except Exception as e:
            raise Exception(f"CausaDB server request failed: {e}")

        data_list = []
        for data_spec in response.get("data", []):
            data = Data(data_spec["name"], self)
            data_list.append(data)

        return data_list



import numpy as np
import pandas as pd


def get_heating_dataset():
    n = 365
    outdoor_temp = 14 + 2 * \
        np.sin(np.linspace(0, 2 * np.pi, n)) + \
        np.random.normal(0, 1, size=n)
    heating = 100 - 3 * outdoor_temp + np.random.normal(0, 1.5, size=n)
    indoor_temp, energy = set_heating(
        heating, outdoor_temp, noise=True)

    outdoor_temp = np.round(outdoor_temp, 2)
    indoor_temp = np.round(indoor_temp, 2)
    energy = np.round(energy, 0)
    heating = np.round(heating, 0)

    data = pd.DataFrame({
        'day': np.arange(n),
        'outdoor_temp': outdoor_temp,
        'heating': heating,
        'indoor_temp': indoor_temp,
        'energy': energy
    })
    return data


def set_heating(heating, outdoor_temp, noise=False):
    indoor_temp = 3 + 0.15 * outdoor_temp + 0.25 * heating
    if noise:
        indoor_temp += np.random.randn(len(outdoor_temp)) * 0.5
    energy = 200 + 2 * indoor_temp + 8 * heating
    if noise:
        energy += np.random.randn(len(outdoor_temp)) * 10
    return indoor_temp, energy


def calculate_wasted_heating_cost(indoor_temp, target_temp, volume=50000, insulation_coefficient=2):
    # 50,000 cubic meters is the size of an average medium commercial warehouse
    # Insulation coefficient of 2 is quite good for a commercial warehouse
    # How far off are we from the target temperature?
    delta_temp = np.abs(indoor_temp - target_temp)
    # kW of heating wasted. 4/3412 is the conversion factor to kWh
    heating_power = volume * delta_temp * insulation_coefficient * (4/3412)
    # Heating power can't be negative
    heating_power = np.maximum(0, heating_power)

    # Daily heating cost
    heating_cost_daily = heating_power * 24 * 0.20  # 20p per kWh (cheap)
    heating_cost_total = np.mean(heating_cost_daily) * 365
    return heating_cost_total, heating_cost_daily, heating_power
